from http_types.utils import HttpExchangeBuilder
from typing import AsyncIterable
import asyncio
from .abstract import AbstractSource
from typing import Tuple

from ..meeshkan_types import HttpExchangeStream
try:
    import faust
except ImportError:
    faust = None


class KafkaProcessorConfig:
    broker: str
    topic: str

    def __init__(self, broker, topic):
        self.broker = broker
        self.topic = topic


class KafkaProcessor:

    def __init__(self, options: KafkaProcessorConfig):
        self.options = options
        self.app = faust.App('myapp', broker=options.broker,
                             stream_wait_empty=False)
        self.faust_topic = self.app.topic(
            options.topic, key_type=str, value_type=str)

        # self.stream = self.faust_topic.stream()

        @self.app.agent(self.faust_topic, sink=[])
        async def gen(recordings: AsyncIterable):
            async for recording in recordings:
                yield HttpExchangeBuilder.from_dict(recording)

        self.agent = gen

    @staticmethod
    def run(app: faust.App, loop: asyncio.AbstractEventLoop, loglevel='info'):
        worker = faust.Worker(app, loop=loop, loglevel=loglevel)

        async def start_worker(worker: faust.Worker) -> None:
            await worker.start()

        try:
            loop.run_until_complete(start_worker(worker))
        finally:
            worker.stop_and_shutdown()


class KafkaSource(AbstractSource):

    def __init__(self, config: KafkaProcessorConfig):

        if faust is None:
            raise Exception(
                "Cannot find module faust. Try `pip install meeshkan[kafka]`")

        self.config = config
        self.processor = KafkaProcessor(config)
        self.recording_stream = self.processor.agent.stream()
        self.worker = None
        self.worker_task = None

    async def start(self, loop: asyncio.AbstractEventLoop) -> Tuple[HttpExchangeStream, asyncio.Task]:
        self.worker = faust.Worker(
            self.processor.app, loop=loop, loglevel='info')

        async def start_worker(worker: faust.Worker) -> None:
            await worker.start()

        source = self.recording_stream
        worker_coro = start_worker(self.worker)
        self.worker_task = loop.create_task(worker_coro)

        return source, self.worker_task

    def shutdown(self):
        self.worker_task.cancel()
        self.worker.stop_and_shutdown()


__all__ = ["KafkaProcessorConfig", "KafkaProcessor", "KafkaSource"]




"""Code for building and inferring media types (application/json, text/plain, etc.) from HTTP exchanges."""
from .update_mode import UpdateMode
from http_types import HttpExchange as HttpExchange
from ..logger import get as getLogger
from typing import Any, Sequence, cast, Optional
from openapi_typed import MediaType, Schema
import json
from typing_extensions import Literal
from .json_schema import to_openapi_json_schema


logger = getLogger(__name__)

MediaTypeKey = Literal['application/json', 'text/plain']

MEDIA_TYPE_FOR_NON_JSON = "text/plain"


def update_json_schema(json_body: Any, mode: UpdateMode, schema: Optional[Any] = None) -> Schema:
    # TODO typeguard
    return cast(Schema, to_openapi_json_schema(json_body, mode, schema))

def nondup(i: Sequence[Any]) -> Sequence[Any]:
    '''Removes duplicates from list'''
    return [json.loads(y) for y in list(set([json.dumps(x) for x in i]))]

def update_text_schema(text_body: str, mode: UpdateMode, schema: Optional[Any] = None) -> Schema:
    # TODO Better updates
    generic = cast(Schema, { "type": "string" })
    specific = cast(Schema, {"type": "string", "enum":[text_body]})
    return generic if mode == UpdateMode.GEN else cast(Schema, {
        "oneOf": nondup([
            specific,
            schema if 'oneOf' not in schema else schema['oneOf']
        ])
    })


def infer_media_type_from_nonempty(body: str) -> MediaTypeKey:
    """Determine media type (application/json, text/plain, etc.) from body.

    Arguments:
        body {str} -- Response body, should not be empty.

    Raises:
        Exception: If body is of unexpected type or empty.

    Returns:
        MediaTypeKey -- Media type such as "application/json"
    """

    if body == '':
        raise Exception("Cannot infer media type from empty body.")

    try:
        as_json = json.loads(body)
    except json.decoder.JSONDecodeError:
        logger.debug(f"Failed decoding: {body}")
        # TODO Handle application/xml etc.
        return MEDIA_TYPE_FOR_NON_JSON

    if isinstance(as_json, dict) or isinstance(as_json, list):
        return 'application/json'
    elif isinstance(as_json, str):
        return 'text/plain'
    else:
        raise Exception(f"Not sure what to do with body: {body}")


def update_media_type(exchange: HttpExchange, mode: UpdateMode, type_key: MediaTypeKey, media_type: Optional[MediaType] = None) -> MediaType:
    """Update media type.

    Arguments:
        exchange {HttpExchange} -- Http exchange
        type_key {MediaTypeKey} -- MediaType such as "application/json"

    Keyword Arguments:
        media_type {Optional[MediaType]} -- Existing media type if exists. For example, "{ 'schema': { 'type': 'string' }}"

    Raises:
        Exception: If existing media type is unknown.

    Returns:
        MediaType -- Updated media type object.
    """
    body = exchange['response']['body']
    existing_schema = (media_type or {}).get("schema", None)

    if type_key == "application/json":
        schema = update_json_schema(json.loads(body), mode, schema=existing_schema)
    elif type_key == "text/plain":
        schema = update_text_schema(body, mode, schema=existing_schema)
    else:
        raise Exception("Unknown type key")
    media_type = MediaType(schema=schema)
    return media_type


def build_media_type(exchange: HttpExchange, mode: UpdateMode, type_key: MediaTypeKey) -> MediaType:
    return update_media_type(exchange=exchange, mode=mode, type_key=type_key, media_type=None)



from functools import reduce
import lenses
from lenses.optics.base import Prism
lens = lenses.lens
import jsonschema
import json
import re
from http_types import Request
from typing import cast, Sequence, Tuple, TypeVar, Callable, Optional, Mapping, Union, Any
from openapi_typed import OpenAPIObject, Response, RequestBody, Header, Operation, Parameter, Components, Reference, Schema, PathItem
from urllib.parse import urlparse

C = TypeVar('C') 
D = TypeVar('D')
S = TypeVar('S')
T = TypeVar('T')
W = TypeVar('W')
X = TypeVar('X')
Y = TypeVar('Y')
Z = TypeVar('Z')

__all__ = ['matcher', 'faker', 'change_ref', 'change_refs']


#def pp(s, c: C) -> C:
#    print(s, c)
#    return c

all_methods: Sequence[str] = [
  "get",
  "post",
  "put",
  "delete",
  "options",
  "head",
  "patch",
  "trace",
]

def omit(p: C, a: str) -> C:
    return { k: v for k, v in p.items() if k != a}

_prism_o = (lambda a: a, lambda b: b)

# TODO are these types correct??
# C because I think the third thingee is the incoming type... maybe?
def odl(focus: Union[str, int]) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    '''Short for "optional-dict-lens, equivalent of Opional in other optics libraries

    Arguments:
        focus {str} -- The key to focus on.
    
    Returns:
        lenses.ui.BaseUiLens[S, T, X, Y] -- A lens that creates the equivalent of an optional
    '''
    return lens.Items().Filter(lambda a: a[0] == focus)[1]
    

def match_urls(protocol: str, host: str, o: OpenAPIObject) -> Sequence[str]:
    """Finds server URLs that match a given protocol and host.

    Arguments:
        protocol {str} -- like http or https
        host {str} -- like api.foo.com
        o {OpenAPIObject} -- schema from which the server URLs are taken
    
    Returns:
        A list of URLs that match the OpenAPI spec.
    """
    return [server['url']
        for server in o['servers']
        if (urlparse(server['url']).scheme == protocol)
            and  (urlparse(server['url']).netloc == host)] if 'servers' in o else []

def get_component_from_ref(o: OpenAPIObject, d: str, accessor: Callable[[Components], Optional[Mapping[str, Union[Reference, C]]]], getter: Callable[[OpenAPIObject, Union[C, Reference]], Optional[C]]) -> Optional[C]:
    return lens.F(
        lambda a: a['components'] if 'components' in a else None
    ).F(
        lambda a: a if a is None else accessor(a)
    ).F(
        lambda a: a if a is None else getter(o, a[d])
    ).get()(o)

def is_reference(a: Any) -> bool:
    return '$ref' in a

def ref_name(r: Reference) -> str:
    return r['$ref'].split("/")[3]

def internal_get_component(f: Callable[[OpenAPIObject, str], Optional[C]]) -> Callable[[OpenAPIObject, Union[C,  Reference]], Optional[C]]:
    def _internal_get_component(o: OpenAPIObject, i: Union[C,  Reference]) -> Optional[C]:
        return (f(o, ref_name(i)) if is_reference(i) else i) if i is not None else None
    return _internal_get_component

# TODO: it seems that there is a bug in pyright that makes get_component_from_ref return a Union[Reference, Schema] instead of Schema
# need to do a cast of the output value to Schema (instead of the union) to fix that
# same for functions below
def get_schema_from_ref(o: OpenAPIObject, d: str) -> Optional[Schema]:
    out = get_component_from_ref(o, d, lambda a: a['schemas'] if 'schemas' in a else None, internal_get_schema_from_ref)
    return None if out is None else cast(Schema, out)

internal_get_schema_from_ref = internal_get_component(get_schema_from_ref)

def get_parameter_from_ref(o: OpenAPIObject, d: str) -> Optional[Parameter]:
    out = get_component_from_ref(o, d, lambda a: a['parameters'] if 'parameters' in a else None, internal_get_parameter_from_ref)
    return None if out is None else cast(Parameter, out)

internal_get_parameter_from_ref = internal_get_component(get_parameter_from_ref)

def get_request_body_from_ref(o: OpenAPIObject, d: str) -> Optional[RequestBody]:
    out = get_component_from_ref(o, d, lambda a: a['requestBodies'] if 'requestBodies' in a else None, internal_get_request_body_from_ref)
    return None if out is None else cast(RequestBody, out)

internal_get_request_body_from_ref = internal_get_component(get_request_body_from_ref)

def get_response_from_ref(o: OpenAPIObject, d: str) -> Optional[Response]:
    out = get_component_from_ref(o, d, lambda a: a['responses'] if 'responses' in a else None, internal_get_response_from_ref)
    return None if out is None else cast(Response, out)

internal_get_response_from_ref = internal_get_component(get_response_from_ref)


def schema_prism(oai: OpenAPIObject) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    return lens.Prism(
        lambda s: get_schema_from_ref(oai, ref_name(s)) if is_reference(s) else s,
        lambda a: a,
        ignore_none=True
    )

def change_ref(j: Reference) -> Reference:
    return {
        '$ref': '#/definitions/%s' % ref_name(j),
    }

def change_refs(j: Schema) -> Schema:
    return cast(Schema, {
        **j,
        **({} if 'additionalProperties' not in j
            else { 'additionalProperties': {} }
            if type(j['additionalProperties']) == type(True)
            else { 'additionalProperties': change_ref(cast(Reference, j['additionalProperties'])) }
            if is_reference(j['additionalProperties'])
            else  { 'additionalProperties': {} }
            if type(j['additionalProperties']) == 'bool'
            else { 'additionalProperties': change_refs(cast(Schema, j['additionalProperties'])) }),
        **({} if 'items' not in j
            else { 'items': change_ref(cast(Reference, j['items'])) }
            if is_reference(j['items'])
            else {
                'items': [change_ref(cast(Reference, item)) if is_reference(item) else change_refs(cast(Schema, item)) for item in cast(Sequence, j['items'])]
            } if type(j['items']) == type([])
            else { 'items': change_refs(cast(Schema, j['items'])) }),
        **({} if 'properties' not in j
            else {
                'properties': reduce(
                    lambda a, b: { **a, b[0]: change_ref(cast(Reference, b[1])) if is_reference(b[1]) else change_refs(cast(Schema, b[1]))},
                    j['properties'].items(),
                    {})
           }),
        **(reduce(
            lambda a, b: { **a, **b },
            [{} if x not in j else { x: [change_ref(cast(Reference, y)) if is_reference(y) else change_refs(cast(Schema, y)) for y in j[x]] } for x in ['anyOf', 'allOf', 'oneOf']],
            {}))
    })

def make_definitions_from_schema(o: OpenAPIObject) -> Mapping[str, Schema]:
    return {
        # TODO: should the second cast be to Union[Schema, Reference]? In unmock it is just Schema, but perhaps it should be both...
        cast(str, a): cast(Schema, change_ref(cast(Reference, b)) if is_reference(b) else change_refs(cast(Schema, b))) for a, b in o['components']['schemas'].items()
    } if ('components' in o) and ('schemas' in o['components']) else {}

def find_relevant_path(m: str, a: Sequence[str], p: PathItem) -> PathItem:
  return p if len(a) == 0 else find_relevant_path(
        m,
        a[1:],
        p if a[0] == m else omit(p, a[0]),
      )


def get_path_item_with_method(m: str, p: PathItem) -> PathItem:
    return find_relevant_path(m, all_methods, p)

def get_required_request_query_or_header_parameters_internal(header: bool, l: lenses.ui.BaseUiLens[S, T, X, Y], oai: OpenAPIObject, p: PathItem) -> Sequence[Parameter]:
    # TODO: really dirty cast
    # is this even the case
    # copied from unmock, but need to investigate
    return cast(Sequence[Parameter], [{ **parameter, 'name': parameter['name'] if not header else parameter['name'].lower(), 'schema': (change_ref(cast(Reference, parameter['schema'])) if is_reference(parameter['schema']) else change_refs(cast(Schema, parameter['schema']))) if 'schema' in parameter else cast(Schema, { 'type': "string" }) } for parameter in l.Prism(
        lambda s: get_parameter_from_ref(oai, ref_name(s)) if is_reference(s) else s,
        lambda a : a,
        ignore_none=True
    ).Prism(
        lambda s : s if (s['in'] == ( "header" if header else "query")) and ('required' in s) and s['required'] else None,
        lambda a : a,
        ignore_none=True
    ).collect()(p) if ('schema' in parameter) and (len(lens.Each().add_lens(schema_prism(oai)).collect()([parameter['schema']])) > 0)])

def get_required_request_query_or_header_parameters(header: bool, req: Request, oai: OpenAPIObject, p: PathItem) -> Sequence[Parameter]:
    return [
        *get_required_request_query_or_header_parameters_internal(
            header,
            odl("parameters").Each(),
            oai,
            p),
        *get_required_request_query_or_header_parameters_internal(
            header,
            odl(req['method'].lower()).add_lens(
                odl("parameters")
            ).Each(),
            oai,
            p)
    ]

def get_required_request_body_schemas(
  req: Request,
  oai: OpenAPIObject,
  p: PathItem,
) -> Sequence[Schema]:
    return odl(req['method'].lower()).add_lens(
        odl("requestBody")
    ).Prism(
        lambda s : get_request_body_from_ref(oai, ref_name(s)) if is_reference(s) else s,
        lambda a: a,
        ignore_none=True
    ).add_lens(
        odl("content")
    ).Values().add_lens(
        odl("schema")
    ).add_lens(schema_prism(oai)).collect()(p)

def valid_schema(to_validate: Any, schema: Any) -> bool:
    try:
        jsonschema.validate(to_validate, schema)
        return True
    except Exception as e:
        # print(e)
        return False

def keep_method_if_required_request_body_is_present(
    req: Request,
    oai: OpenAPIObject,
) -> Callable[[PathItem], PathItem]:
    def _keep_method_if_required_request_body_is_present(p: PathItem) -> PathItem:
        out = p if (req['method'].lower() not in p) or (len([s for s in get_required_request_body_schemas(req, oai, p) if
                not valid_schema(req['bodyAsJson'], {
                    # TODO: this line is different than the TS implementation
                    # because I think there is a logic bug there
                    # it should look like this line as we are not sure
                    # if the schema will be a reference or not
                    # perhaps I'm wrong in the assumption... only testing will tell...
                    # otherwise, change the TS implementation in unmock-js and delete this comment.
                    **(change_ref(cast(Reference, s)) if is_reference(s) else change_refs(cast(Schema, s))),
                    'definitions': make_definitions_from_schema(oai),
                })]) == 0) else omit(p, req['method'].lower())
        return out
    return _keep_method_if_required_request_body_is_present

def keep_method_if_required_header_parameters_are_present(
    req: Request,
    oai: OpenAPIObject,
) -> Callable[[PathItem], PathItem]:
    def _keep_method_if_required_header_parameters_are_present(p: PathItem) -> PathItem:
        out = keep_method_if_required_query_or_header_parameters_are_present(True, req, oai, p)
        return out
    return _keep_method_if_required_header_parameters_are_present

def keep_method_if_required_query_parameters_are_present(
    req: Request,
    oai: OpenAPIObject,
) -> Callable[[PathItem], PathItem]:
    def _keep_method_if_required_query_parameters_are_present(p: PathItem) -> PathItem:
        return keep_method_if_required_query_or_header_parameters_are_present(False, req, oai, p)
    return _keep_method_if_required_query_parameters_are_present

def _json_schema_from_required_parameters(parameters: Sequence[Parameter], oai: OpenAPIObject) -> Any:
    return {
        'type': 'object',
        'properties': {param['name']: param['schema'] for param in parameters},
        'required': [param['name'] for param in parameters if ('required' in param) and param['required']],
        'additionalProperties': True,
        'definitions': make_definitions_from_schema(oai),
    }


def keep_method_if_required_query_or_header_parameters_are_present(
    header: bool,
    req: Request,
    oai: OpenAPIObject,
    p: PathItem
) -> PathItem:
    return p if (req['method'].lower() not in p) or valid_schema(
        { k.lower(): v for k, v in req['headers'].items() } if header else req['query'],
        _json_schema_from_required_parameters(get_required_request_query_or_header_parameters(header, req, oai, p), oai)
    ) else omit(p, req['method'].lower())

def maybe_add_string_schema(s: Sequence[Union[Reference, Schema]]) -> Sequence[Union[Reference, Schema]]:
    return [cast(Schema, { 'type': "string" })] if len(s) == 0 else s

def discern_name(o: Optional[Parameter], n: str) -> Optional[Parameter]:
    return o if (o is None) or ((o['name'] == n) and (o['in'] == 'path')) else None

def internal_get_parameter_schemas(
    t: lenses.ui.BaseUiLens[S, T, X, Y],
    vname: str,
    path_item: PathItem,
    oas: OpenAPIObject,
) -> Sequence[Schema]:
    return t.Prism(
        lambda i: discern_name(get_parameter_from_ref(oas, ref_name(i)), vname) if is_reference(i) else discern_name(i, vname),
        lambda a: a,
        ignore_none=True
    ).add_lens(
        odl("schema")
    ).collect()(path_item)


def path_item_path_parameter_schemas(
    vname: str,
    path_item: PathItem,
    oas: OpenAPIObject,
) -> Sequence[Schema]:
    return internal_get_parameter_schemas(
        odl('parameters').Each(),
        vname,
        path_item,
        oas,
    )

def operation_path_parameter_schemas(
  vname: str,
  path_item: PathItem,
  operation: str,
  oas: OpenAPIObject,
) -> Sequence[Schema]:
    return internal_get_parameter_schemas(
        odl(
            operation.lower()
        ).add_lens(
            odl('parameters')
        ).Each(),
        vname,
        path_item,
        oas,
  )

def get_matching_parameters(
  vname: str,
  path_item: PathItem,
  operation: str,
  oas: OpenAPIObject,
) -> Sequence[Union[Schema, Reference]]:
    return maybe_add_string_schema([
        *path_item_path_parameter_schemas(vname, path_item, oas),
        *operation_path_parameter_schemas(vname, path_item, operation, oas)
    ])

def maybeJson(maybe: str) -> Any:
  try:
    return json.loads(maybe)
  except:
    return maybe

def path_parameter_match(
    part: str,
    vname: str,
    path_item: PathItem,
    operation: str,
    oas: OpenAPIObject,
) -> bool:
    '''Matches part of a path against a path parameter with name vname
    
    Arguments:
        part {str} -- part of a path, ie an id
        vname {str} -- name of a parameter
        path_item {PathItem} -- a path item maybe containing the parameter
        operation {MethodNames} -- the name of the operation to check in case the parameter is in the operation
        oas {OpenAPIObject} -- the schema to traverse to find definitions
    
    Returns:
        bool -- Did we get a match
    '''
    return reduce(lambda q, r: q and reduce(
        lambda a, b:
          a or
          valid_schema(b, {
            **r,
            'definitions': make_definitions_from_schema(oas),
          }),
        list(set([part, maybeJson(part)])),
        False,
    ), get_matching_parameters(vname, path_item, operation, oas), True)

path_param_regex = re.compile(r'^\{[^}]+\}')

# TODO: this is a boolean nightmare
# what we want is
# lengths_are_equal AND (empty OR ((first_elt_equal OR (matches_regex AND matches_schema)) AND recursion))
def matches_internal(
    path: Sequence[str],
    path_item_key: Sequence[str],
    path_item: PathItem,
    operation: str,
    o: OpenAPIObject,
) -> bool:
    return (len(path) == len(path_item_key)) and (
        (len(path) == 0) or
        (((path[0] == path_item_key[0]) or
        ((path_param_regex.match(path_item_key[0]) is not None) and
            path_parameter_match(
                path[0],
                path_item_key[0][1:-1],
                path_item,
                operation,
                o,
            ))) and
            matches_internal(
                path[1:],
                path_item_key[1:],
                path_item,
                operation,
                o,
            )))

def matches(
  path: str,
  path_item_key: str,
  path_item: PathItem,
  method: str,
  oas: OpenAPIObject,
) -> bool:
  return matches_internal(
    [x for x in path.split("/") if x != ""],
    [x for x in path_item_key.split("/") if x != ""],
    path_item,
    method,
    oas,
  )

def get_first_method_internal_2(
  p: PathItem,
  n: str,
  m: Sequence[str],
  o: Optional[Operation],
) ->  Optional[Tuple[str, Operation]]:
    return (n, o) if o is not None else get_first_method_internal(m, p)

def get_first_method_internal(
  m: Sequence[str],
  p: PathItem,
) -> Optional[Tuple[str, Operation]]:
    return None if len(m) == 0 else get_first_method_internal_2(p, m[0], m[1:], p[m[0]] if m[0] in p else None)

def get_first_method(p: PathItem) -> Optional[Tuple[str, Operation]]:
  return get_first_method_internal(all_methods, p)

def get_header_from_ref(
    o: OpenAPIObject,
    d: str
) -> Optional[Header]:
    out = get_component_from_ref(
        o,
        d,
        lambda a: a['headers'] if 'headers' in a else None,
        internal_get_header_from_ref,
    )
    return None if out is None else cast(Header, out)

internal_get_header_from_ref = internal_get_component(get_header_from_ref)

def use_if_header_last_mile(
  p: Parameter,
  r: Optional[Schema],
) -> Optional[Tuple[str, Schema]]:
  return None if r is None else (p['name'], r if r is not None else cast(Schema, { 'type': "string" }))

def use_if_header(
  o: OpenAPIObject,
  p: Parameter,
) -> Optional[Tuple[str, Schema]]:
    return None if p['in'] != "header" else use_if_header_last_mile(
        p,
        cast(Schema, { 'type': "string" }) if ('schema' not in p) or (p['schema'] is None) else
            get_schema_from_ref(o, ref_name(cast(Reference, p['schema']))) if is_reference(p['schema']) else
                cast(Schema, p['schema'])
    )

def parameter_schema(o: OpenAPIObject) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    return lens.Iso(lambda a: (use_if_header(o, a), a), lambda b: b[1])[0].Prism(*_prism_o, ignore_none=True)


def cut_path(paths: Sequence[str], path: str) -> str:
    return path if len(paths) == 0 else path[len(paths[0]):] if path[:len(paths[0])] == paths[0] else cut_path(paths[1:], path)

def remove_trailing_slash(s: str) -> str:
    return s if len(s) == 0 else s[:-1] if s[-1] == "/" else s

def truncate_path(
  path: str,
  o: OpenAPIObject,
  i: Request,
) -> str:
    return cut_path(
        [remove_trailing_slash(urlparse(u).path) for u in match_urls(i['protocol'], i['host'], o)],
        path,
    )

def matcher(req: Request, r: Mapping[str, OpenAPIObject]) -> Mapping[str, OpenAPIObject]:
    return lens.Values().modify(
        lambda oai: odl("paths").Values().modify(
            lambda path_item: reduce(lambda a, b: b(a), [
                keep_method_if_required_header_parameters_are_present(req, oai),
                keep_method_if_required_query_parameters_are_present(req, oai),
                keep_method_if_required_request_body_is_present(req, oai),
            ], get_path_item_with_method(req['method'].lower(), path_item))
        )({
            **oai,
            **({} if 'paths' not in oai else {
              'paths': {n: o for n, o in oai['paths'].items() if matches(
                    truncate_path(req['pathname'], oai, req),
                    n,
                    o,
                    req['method'].lower(),
                    oai,
                  )}
                })
        })
    )({k: v for k, v in r.items() if len(match_urls(req['protocol'], req['host'], v)) > 0 })

def _first_or_none(l: Sequence[C]) -> Optional[C]:
    return None if len(l) == 0 else l[0]

def is_parameter(s: Any) -> bool:
    return s is not None and 'in' in s and 'name' in s

def headers_schemas_from_operation(schema: OpenAPIObject, operation: Operation) -> Sequence[Schema]:
    return odl("parameters").Each().Prism(
        lambda s: s if is_parameter(s) else get_parameter_from_ref(schema, ref_name(s)) if is_reference(s) else None,
        lambda a: a,
        ignore_none=True
    ).Iso(
        lambda a: (use_if_header(schema, a), a),
        lambda b: b[1]
    )[0].Prism(*_prism_o, ignore_none=True).collect()(operation)

# TODO: this is pretty impovrished compared to unmock JS
# can we make it better
def is_response(a: Any) -> bool:
    return (a is not None) and ('description' in a) and (True if 'content' not in a else type(a['content']) == type({}))

def make_lens_to_response_starting_from_operation(
    schema: OpenAPIObject,
    code: str,
    l: lenses.ui.BaseUiLens[S, T, X, Y] = lens
) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    return l.add_lens(odl("responses")).Prism(
        lambda s: s if is_response(s) else 
            get_response_from_ref(schema, ref_name(s)) if is_reference(s) else
            None,
        lambda a: a,
        ignore_none=True
    )

def string_header(n: str, o: Optional[Header]) -> Optional[Tuple[str, Header]]:
    return None if o is None else (n, o)

def header_schemas_from_response(
  schema: OpenAPIObject,
  operation: Operation,
  code: str
) -> Sequence[Schema]:
    return make_lens_to_response_starting_from_operation(
      schema, code
    ).add_lens(odl("headers")).Items().Prism(
      lambda a: string_header(a[0], get_header_from_ref(schema, ref_name(a[1]))) if is_reference(a[1]) else (a[0], a[1]),
      lambda a: a,
      ignore_none=True
    ).Iso(
      lambda a: { **a[1], 'name': a[0], 'in': 'header' },
      lambda a: (a['name'], a)
    ).add_lens(parameter_schema(schema)).collect()(operation)

def body_from_response(
  schema: OpenAPIObject,
  operation: Operation,
  code: str
) -> Optional[Union[Reference, Schema]]:
  return _first_or_none(
    make_lens_to_response_starting_from_operation(
          schema, code
    ).add_lens(
        odl("content")
    ).Items().add_lens(
        odl(0)
    )[1].add_lens(
        odl("schema")
    ).collect()([operation]))


########################
#### FAKER
########################
import random
from faker import Faker
fkr = Faker()
_LO = -99999999
_HI = 99999999

# to prevent too-nested objects
def sane_depth(n):
    return max([0, 3-n])

def fake_object(schema: Any, top_schema: Any, depth: int) -> Any:
    addls = {} if 'additionalProperties' not in schema else {k:v for k,v in [(fkr.name(), random.random() if (type(schema['additionalProperties']) == type(True)) and (schema['additionalProperties'] == True) else faker(schema['additionalProperties'], top_schema, depth)) for x in range(random.randint(0, 4))]}
    properties = [] if 'properties' not in schema else [x for x in schema['properties'].keys()]
    random.shuffle(properties)
    properties = [] if len(properties) == 0 else properties[:min([sane_depth(depth), random.randint(0, len(properties) - 1 )])]
    properties = list(set(([] if 'required' not in schema else schema['required']) + properties))
    return {
        **addls,
        **{ k: v for k,v in [(p, faker(schema['properties'][p], top_schema, depth)) for p in properties]}
    }

def fake_array(schema: Any, top_schema: Any, depth: int) -> Any:
    mn = 0 if 'minItems' not in schema else schema['minItems']
    mx = 100 if 'maxItems' not in schema else schema['maxItems']
    return [faker(x, top_schema, depth) for x in schema['items']] if type(schema['items']) is type([]) else [faker(schema['items'], top_schema, depth) for x in range(random.randint(mn, mx))]

def fake_any_of(schema: Any, top_schema: Any, depth: int) -> Any:
    return faker(random.choice(schema["anyOf"]), top_schema, depth)

def fake_all_of(schema: Any, top_schema: Any, depth: int) -> Any:
    return reduce(lambda a, b: { **a, **b}, [faker(x, top_schema, depth) for x in schema["allOf"]], {})

def fake_one_of(schema: Any, top_schema: Any, depth: int) -> Any:
    return faker(random.choice(schema["oneOf"]), top_schema, depth)

# TODO - make this work
def fake_not(schema: Any, top_schema: Any, depth: int) -> Any:
    return {}

# TODO - make this not suck
def fake_string(schema: Any) -> str:
    return random.choice(schema['enum']) if 'enum' in schema else fkr.name()

def fake_boolean(schema: Any) -> bool:
    return random.choice(schema['enum']) if 'enum' in schema else True if random.random() > 0.5 else False

# TODO: add exclusiveMinimum and exclusiveMaximum
def fake_integer(schema: Any) -> int:
    mn = _LO if 'minimum' not in schema else schema['minimum']
    mx = _HI if 'maximum' not in schema else schema['maximum']
    return random.choice(schema['enum']) if 'enum' in schema else random.randint(mn, mx)

def fake_ref(schema: Any, top_schema, depth):
    name = schema['$ref'].split('/')[2]
    return faker(top_schema['definitions'][name], top_schema, depth)

def fake_null(schema: Any) -> None:
    return None

def fake_number(schema: Any) -> float:
    mn = _LO if 'minimum' not in schema else schema['minimum']
    mx = _HI if 'maximum' not in schema else schema['maximum']
    return random.choice(schema['enum']) if 'enum' in schema else (random.random() * (mx - mn)) + mn

def faker(schema: Any, top_schema: Any, depth: int) -> Any:
    depth += 1
    return fake_array(
        schema, top_schema, depth
    ) if ('type' in schema) and (schema["type"] == "array") else fake_any_of(
        schema, top_schema, depth
    ) if "anyOf" in schema else fake_all_of(
        schema, top_schema, depth
    ) if "allOf" in schema else fake_one_of(
        schema, top_schema, depth
    ) if "oneOf" in schema else fake_not(
        schema, top_schema, depth
    ) if "not" in schema else fake_ref(
        schema, top_schema, depth
    ) if "$ref" in schema else fake_object(
        schema, top_schema, depth
    ) if ("type" not in schema) or (schema["type"] == "object") else fake_string(
        schema
    ) if schema["type"] == "string" else fake_integer(
        schema
    ) if schema["type"] == "integer" else fake_boolean(
        schema
    ) if schema["type"] == "boolean" else fake_null(
        schema
    ) if schema["type"] == "null" else fake_number(
        schema
    ) if schema["type"] == "number" else {}


import abc
import asyncio
from typing import Tuple, Optional
from ..meeshkan_types import HttpExchangeStream


class AbstractSource(abc.ABC):
    """Abstract source of HTTP exchanges.
    """

    @abc.abstractmethod
    async def start(self, loop: asyncio.AbstractEventLoop) -> Tuple[HttpExchangeStream, Optional[asyncio.Task]]:
        """Start source task to read HTTP Exchanges.

        Arguments:
            loop {asyncio.AbstractEventLoop} -- Event loop to use.

        Returns:
            Tuple[HttpExchangeStream, Optional[asyncio.Task]] -- Stream of HttpExchange objects and possibly a reader task.
        """
        raise NotImplementedError("")

    @abc.abstractmethod
    def shutdown(self) -> None:
        """Shutdown source.
        """
        raise NotImplementedError("")



import json
import logging
from meeshkan.schemabuilder.update_mode import UpdateMode
import os

from typing import cast
import yaml
from http_types import Request, Response, HttpExchange, RequestBuilder, ResponseBuilder

from ...schemabuilder.builder import BASE_SCHEMA, update_openapi

logger = logging.getLogger(__name__)


class DataCallback:
    def log(self, request: Request, response: Response):
        raise NotImplementedError()


class RequestLoggingCallback():
    _log_dir: str
    _schema_dir: str
    def __init__(self, recording=False, log_dir=None, schema_dir=None, append=True):
        self._recording = recording
        self._append = append
        self._log_dir = log_dir
        self._schema_dir = schema_dir

        self._logs = {}
        self._schemas = {}

        if self._recording:
            if not os.path.exists(self._log_dir):
                os.makedirs(self._log_dir)
            if not os.path.exists(self._schema_dir):
                os.makedirs(self._schema_dir)

    def log(self, request: Request, response: Response):
        RequestBuilder.validate(request)
        ResponseBuilder.validate(response)

        host = request['headers']['Host']
        reqres = HttpExchange(request=request, response=response)
        if not host in self._logs:
            log_file = os.path.join(self._log_dir, '{}.jsonl'.format(host))
            if self._append and os.path.exists(log_file):
                self._logs[host] = open(log_file, 'a')
            else:
                self._logs[host] = open(log_file, 'w')

        self._logs[host].write(json.dumps(reqres))
        self._logs[host].write('\n')

        schema_dir = os.path.join(self._schema_dir, cast(str, host))
        if not os.path.exists(schema_dir):
            os.makedirs(schema_dir)

        schema_file = os.path.join(schema_dir, 'openapi.yaml')

        if not host in self._schemas:
            if os.path.exists(schema_file) and self._append:
                with open(schema_file, 'r') as f:
                    self._schemas[host] = yaml.load(f)
            else:
                self._schemas[host] = BASE_SCHEMA

        # TODO: is hardcoded update mode correct?
        self._schemas[host] = update_openapi(self._schemas[host], reqres, UpdateMode.GEN)

        with open(schema_file, 'w') as f:
            yaml.dump(self._schemas[host], f)

        logger.debug('Updated logs and scheme for host %s', host)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for f in self._logs.values():
            f.close()



import glob
import importlib.util
import inspect
import json
import logging
import os

from .storage import storage_manager

logger = logging.getLogger(__name__)


class CallbackManager:
    ARGS_MAP_COMMON = {
        'query': lambda request, response, storage: request['query'],
        'request_headers': lambda request, response, storage: request['headers'],
        'respone_headers': lambda request, response, storage: response['headers'],
        'request': lambda request, response, storage: request,
        'response': lambda request, response, storage: response,
        'storage': lambda request, response, storage: storage,
    }

    ARGS_MAP_JSON = {**ARGS_MAP_COMMON,
                     'request_body': lambda request, response, storage: request['bodyAsJson'],
                     'response_body': lambda request, response, storage: response['bodyAsJson']
                     }

    ARGS_MAP_TEXT = {**ARGS_MAP_COMMON,
                     'request_body': lambda request, response, storage: request['body'],
                     'response_body': lambda request, response, storage: response['body']
                     }

    def __init__(self):
        self._callbacks = dict()

    def load(self, path):
        if not os.path.exists(path):
            logger.debug("Callback configuration path doesn't exist %s", path)

        for f in glob.glob(os.path.join(path, '*.py')):
            module_name = 'callbacks.{}'.format(os.path.splitext(os.path.basename(f))[0])
            logging.debug('Loading callbacks from %s to %s module', f, module_name)
            # from https://stackoverflow.com/questions/19009932/import-arbitrary-python-source-file-python-3-3
            spec = importlib.util.spec_from_file_location(module_name, f)
            callbacks = importlib.util.module_from_spec(spec)
            # even though this is an accepted response, the line below
            # triggers a typing error. ignore types for now, but should
            # investigate why the type does not work
            spec.loader.exec_module(callbacks) # type: ignore

        logging.info('Loaded %d callbacks', len(self._callbacks))

    def callback(self, host, method, path):
        return self._callbacks.get((host, method, path))

    def add_callback(self, host, method, path, format, response_type, callback):
        args = inspect.getfullargspec(callback).args
        args_map = self.ARGS_MAP_JSON if format == 'json' else self.ARGS_MAP_TEXT
        self._callbacks[(host, method, path)] = lambda request, response, storage: \
            self._map_results(response, callback(**{arg: args_map[arg](request, response, storage) for arg in args}),
                              format, response_type)

    def _map_results(self, response, result, format, response_type):
        if response_type == 'body':
            if format == 'json':
                response['bodyAsJson'] = result
                response['body'] = json.dumps(result)
            else:
                response['body'] = result
        else:
            if format == 'json':
                response['body'] = json.dumps(result)

        return response

    def __call__(self, request, response):
        callback = self._callbacks.get((request['host'], request['method'], request['pathname']))
        if callback is not None:
            return callback(request, response, storage_manager.default)
        else:
            return response


callback_manager = CallbackManager()


def callback(host, method, path, format='json', response='body'):
    def decorator(function):
        callback_manager.add_callback(host, method, path, format, response, function)
        return function
    return decorator








import copy
from .update_mode import UpdateMode
from functools import reduce
from typing import Any, List, Sequence, Iterable, AsyncIterable, cast, Tuple, Optional, Union, TypeVar, Type, Mapping
from urllib.parse import urlunsplit
from collections import defaultdict
import json

from http_types import HttpExchange as HttpExchange
from openapi_typed import Info, Header, MediaType, OpenAPIObject, PathItem, Response, Operation, Parameter, Reference, Server, Responses
from typeguard import check_type  # type: ignore

from ..logger import get as getLogger
from .media_types import infer_media_type_from_nonempty, build_media_type, update_media_type, MediaTypeKey, update_text_schema
from .paths import find_matching_path, RequestPathParameters
from .param import ParamBuilder
from .schema import validate_openapi_object
from .servers import normalize_path_if_matches
from .result import BuildResult


logger = getLogger(__name__)

__all__ = ['build_schema_batch', 'build_schema_online', 'update_openapi']

def build_response_content(exchange: HttpExchange, mode: UpdateMode) -> Optional[Tuple[MediaTypeKey, MediaType]]:
    """Build response content schema from exchange.

    Arguments:
        request {HttpExchange} -- Http exchange.

    Returns:
        Optional[Tuple[str, MediaType]] -- None for empty body, tuple of media-type key and media-type otherwise.
    """
    body = exchange['response']['body']

    if body == '':
        return None

    media_type_key = infer_media_type_from_nonempty(body)

    media_type = build_media_type(exchange, mode, type_key=media_type_key)

    return (media_type_key, media_type)


def build_response(request: HttpExchange, mode: UpdateMode) -> Response:
    """Build new response object from request response pair.

    Response reference: https://swagger.io/specification/#responseObject

    Arguments:
        request {HttpExchange} -- Request-response pair.

    Returns:
        Response -- OpenAPI response object.
    """
    # TODO Headers and links
    content_or_none = build_response_content(request, mode)

    if content_or_none is None:
        return Response(
            description="Response description",
            headers={},
            links={}
        )

    media_type_key, media_type = content_or_none
    content = {
        media_type_key: media_type
    }

    return Response(
        description="Response description",
        headers={},
        content=content,
        links={}
    )


def update_response(response: Response, mode: UpdateMode, exchange: HttpExchange) -> Response:
    """Update response object. Mutates the input object.

    Response reference: https://swagger.io/specification/#responseObject

    Arguments:
        response {Response} -- Existing response object.
        exchange {HttpExchange} -- Request-response pair.

    Returns:
        Response -- Updated response object.
    """
    # TODO Update headers and links
    response_body = exchange['response']['body']

    if response_body == '':
        # No body, do not do anything
        # TODO How to mark empty body as a possible response if non-empty responses exist
        return response

    media_type_key = infer_media_type_from_nonempty(response_body)

    ################
    ### HEADERS
    ################
    response_headers = response.get('headers', None)
    useable_headers = {} if response_headers is None else { k: v for k,v in cast(Mapping[str, str], response_headers).items() if k not in ['content-type', 'content-length', 'Content-Type', 'Content-Length']}
    has_headers = len(useable_headers) > 0
    if response_headers is None and has_headers:
        response['headers'] = {}
    if has_headers:
        response['headers'] = {
            **response['headers'],
            **{ k: update_text_schema(v, mode, schema=response['headers'].get(k, None)) for k, v in useable_headers.items() }
        }

    #############
    ### CONTENT
    #############
    response_content = response['content'] if 'content' in response else None
    if response_content is not None and media_type_key in response_content:
        # Need to update existing media type
        existing_media_type = response_content[media_type_key]
        media_type = update_media_type(
            exchange=exchange, mode=mode, type_key=media_type_key, media_type=existing_media_type)
    else:
        media_type = build_media_type(
            exchange=exchange, mode=mode, type_key=media_type_key)

    response_content = {**response_content, **{media_type_key: media_type}}
    response['content'] = response_content
    return response


def build_operation(exchange: HttpExchange, mode: UpdateMode) -> Operation:
    """Build new operation object from request-response pair.

    Operation reference: https://swagger.io/specification/#operationObject

    Arguments:
        request {HttpExchange} -- Request-response pair

    Returns:
        Operation -- Operation object.
    """
    response = build_response(exchange, mode)
    code = str(exchange['response']['statusCode'])

    request_query_params = exchange['request'].get('query', {})
    request_header_params = exchange['request'].get('headers', {})
    schema_query_params = ParamBuilder('query').build(request_query_params, mode)
    # TODO: unfreeze
    schema_header_params = ParamBuilder('header').build(request_header_params, mode)

    operation = Operation(
        summary="Operation summary",
        description="Operation description",
        operationId="id",
        responses={code: response},
        parameters=[*schema_query_params, *schema_header_params])
    return operation


def update_operation(operation: Operation, request: HttpExchange, mode: UpdateMode) -> Operation:
    """Update OpenAPI operation object. Mutates the input object.

    Operation reference: https://swagger.io/specification/#operationObject

    Arguments:
        operation {Operation} -- Existing Operation object.
        request {HttpExchange} -- Request-response pair

    Returns:
        Operation -- Updated operation
    """
    responses = operation['responses']  # type: Responses
    response_code = str(request['response']['statusCode'])
    if response_code in responses:
        # Response exists
        existing_response = responses[response_code]
        # Ensure response is Response and not Reference
        # check_type('existing_response', existing_response, Response)
        existing_response = cast(Response, existing_response)
        response = update_response(existing_response, mode, request)
    else:
        response = build_response(request, mode)

    existing_parameters = operation['parameters']
    request_query_params = request['request'].get('query', {})
    request_header_params = request['request'].get('header', {})
    ## HACK
    # because update returns a full list, we need to merge
    # we do that using json
    updated_parameters = [json.loads(y) for y in set([json.dumps(x) for x in [*ParamBuilder('query').update(request_query_params,
            mode, existing_parameters),
            # TODO: can we avoid the cast below? it is pretty hackish
        *ParamBuilder('header').update(cast(Mapping[str, Union[str, Sequence[str]]], request_header_params),
            mode, existing_parameters)]])]

    operation['parameters'] = updated_parameters
    operation['responses'][response_code] = response
    return operation


T = TypeVar('T')


def verify_not_ref(item: Union[Reference, Any], expected_type: Type[T]) -> T:
    assert not '$ref' in item, "Did not expect a reference"
    return cast(T, item)


def update_openapi(schema: OpenAPIObject, request: HttpExchange, mode: UpdateMode) -> OpenAPIObject:
    """Update OpenAPI schema with a new request-response pair.
    Does not mutate the input schema.

    Schema reference: https://swagger.io/specification/#oasObject

    Returns:
        OpenAPI -- Updated schema
    """
    schema_copy = copy.deepcopy(schema)
    request_method = request['request']['method']
    request_path = request['request']['pathname']

    if not 'servers' in schema_copy:
        schema_copy['servers'] = []
    schema_servers = schema_copy['servers']

    normalized_pathname_or_none = normalize_path_if_matches(
        request['request'], schema_servers)
    if normalized_pathname_or_none is None:
        schema_copy['servers'] = [*schema_copy['servers'], Server(url=urlunsplit(
            [request['request']['protocol'], request['request']['host'], '', '', '']))]
        normalized_pathname = request_path
    else:
        normalized_pathname = normalized_pathname_or_none

    schema_paths = schema_copy['paths']
    operation_candidate = build_operation(request, mode)
    path_match_result = find_matching_path(normalized_pathname, schema_paths, request_method, operation_candidate)

    request_path_parameters = {}

    if path_match_result is not None:
        # Path item exists for request path
        path_item, request_path_parameters, pathname_with_wildcard, pathname_to_be_replaced_with_wildcard = path_match_result['path'], path_match_result['param_mapping'], path_match_result['pathname_with_wildcard'], path_match_result['pathname_to_be_replaced_with_wildcard']
        # Create a wildcard if the mode is not replay/mixed
        if (pathname_to_be_replaced_with_wildcard is not None):
            if mode == UpdateMode.GEN:
                # the algorithm has updated the pathname, need to mutate
                # the schema paths to use the new and discard the old if the old exists
                # in the schema. it would not exist if we have already put a wildcard
                pointer_to_value = schema_paths[pathname_to_be_replaced_with_wildcard]
                schema_paths = { k: v for k, v in [(pathname_with_wildcard, pointer_to_value), *schema_paths.items()] if k != pathname_to_be_replaced_with_wildcard }
                if not ('parameters' in schema_paths[pathname_with_wildcard].keys()):
                    schema_paths[pathname_with_wildcard]['parameters'] = []
                for path_param in request_path_parameters.keys():
                    params = [cast(Parameter, x) for x in schema_paths[pathname_with_wildcard]['parameters'] if '$ref' not in x]
                    if not (path_param in [x['name'] for x in params if x['in'] == 'path']):
                        schema_paths[pathname_with_wildcard]['parameters'] = [{
                            'required': True,
                            'in': 'path',
                            'name': path_param,
                        }, *(schema_paths[pathname_with_wildcard]['parameters'])]
                schema_copy['paths'] = schema_paths
            else:
                # we are using recordings, so we shouldn't overwrite anything
                # we only add if it is not there yet
                if request_path not in schema_paths:
                    # TODO: merge with liens below?
                    path_item = PathItem(summary="Path summary",
                             description="Path description")
                    request_path_parameters = {}
                    schema_copy['paths'] = {**schema_paths, **{request_path: path_item}}
    else:
        path_item = PathItem(summary="Path summary",
                             description="Path description")
        request_path_parameters = {}
        schema_copy['paths'] = {**schema_paths, **{request_path: path_item}}
    if request_method in path_item:
        # Operation exists
        existing_operation = path_item[request_method]  # type: ignore
        operation = update_operation(existing_operation, request, mode)

    else:
        operation = operation_candidate

    # Needs type ignore as one cannot set variable property on typed dict
    path_item[request_method] = operation  # type: ignore
    return cast(OpenAPIObject, schema_copy)


BASE_SCHEMA = OpenAPIObject(openapi="3.0.0",
                            info=Info(title="API title",
                                      description="API description", version="1.0"),
                            paths={})


async def build_schema_async(async_iter: AsyncIterable[HttpExchange],  mode: UpdateMode, starting_spec: OpenAPIObject) -> AsyncIterable[BuildResult]:
    schema = starting_spec
    async for exchange in async_iter:
        schema = update_openapi(schema, exchange, mode)
        yield BuildResult(openapi=schema)


def build_schema_online(requests: Iterable[HttpExchange], mode: UpdateMode, base_schema: OpenAPIObject = BASE_SCHEMA) -> OpenAPIObject:
    """Build OpenAPI schema by iterating request-response pairs.

    OpenAPI object reference: https://swagger.io/specification/#oasObject

    Arguments:
        requests {Iterator[HttpExchange]} -- Iterator of request-response pairs.

    Returns:
        OpenAPI -- OpenAPI object.
    """

    # Iterate over all request-response pairs in the iterator, starting from
    # BASE_SCHEMA
    schema = reduce(lambda schema, req: update_openapi(
        schema, req, mode), requests, base_schema)

    validate_openapi_object(schema)

    return schema


def build_schema_batch(requests: List[HttpExchange], mode: UpdateMode, base_schema: OpenAPIObject = BASE_SCHEMA) -> OpenAPIObject:
    """Build OpenAPI schema from a list of request-response object.

    OpenAPI object reference: https://swagger.io/specification/#oasObject

    Arguments:
        requests {List[HttpExchange]} -- List of request-response pairs.

    Returns:
        OpenAPI -- OpenAPI object.
    """
    return build_schema_online(iter(requests), mode, base_schema=base_schema)



import tornado.tcpserver
from http_types import Request, Response


class ProxyCallback(tornado.tcpserver.TCPServer):
    def __init__(self, ssl_options: bool = None):
        super().__init__(ssl_options=ssl_options)

    def on_remove_channel(self, client_address: tuple):
        raise NotImplementedError()

    def on_request_complete(self, request: Request, response: Response):
        raise NotImplementedError()



import copy
import glob
import json
import logging
import os
import yaml
import random
from faker import Faker
from typing import cast, Mapping, Union, Sequence, Tuple
from openapi_typed import OpenAPIObject, Reference, Schema
from ...gen.generator import match_urls, matcher, faker, change_ref, change_refs
from http_types import Response, HttpMethod
fkr = Faker()

from http_types import Request, Response

logger = logging.getLogger(__name__)

class ResponseMatcher:
    _schemas: Mapping[str, OpenAPIObject]
    def __init__(self, schema_dir):
        schemas: Sequence[str] = []
        if not os.path.exists(schema_dir):
            logging.info('OpenAPI schema directory not found %s', schema_dir)
        else:
            schemas = [s for s in os.listdir(schema_dir) if s.endswith('yml') or s.endswith('yaml')]
        specs: Sequence[Tuple[str, OpenAPIObject]] = []
        for schema in schemas:
            with open(os.path.join(schema_dir, schema), encoding='utf8') as schema_file:
                # TODO: validate schema?
                specs = [*specs, (schema, cast(OpenAPIObject, yaml.safe_load(schema_file.read())))]
        self._schemas = { k: v for k, v in specs}

    def match_error(self, msg: str, req: Request) -> Response:
        return self.default_response('%s. Here is the full request: host=%s, path=%s, method=%s.' % (msg, req['host'], req['path'], req['method']))

    def default_response(self, msg):
        json_resp = {'message': msg}
        return Response(statusCode=500, body=json.dumps(json_resp), bodyAsJson=json_resp,
                        headers={})


    def get_response(self, request: Request) -> Response:
        # mutate schema to include current hostname in servers
        # this is a hack - should be more versatile than this in future
        match = matcher(request, self._schemas)
        if len(match) == 0:
            return self.match_error('Could not find a open API schema for the host %s.' % request['host'], request)
        match_keys = [x for x in match.keys()]
        random.shuffle(match_keys)
        name = match_keys[0]
        path_error = 'Could not find a path %s on hostname %s.' % (request['path'], request['host'])
        method_error = 'Could not find a method %s for path %s on hostname %s.' % (request['method'], request['path'], request['host'])
        if 'paths' not in match[name]:
            return self.match_error(path_error, request)
        if len(match[name]['paths'].items()) == 0:
            return self.match_error(path_error, request)
        if request['method'] not in [x for x in match[name]['paths'].values()][0].keys():
            return self.match_error(method_error, request)
        method = [x for x in match[name]['paths'].values()][0][cast(HttpMethod, request['method'])]
        responses_error = 'While a stub for a specification exists for this endpoint, it contains no responses. That usually means the schema is corrupt or it has been constrained too much (ie asking for a 201 response when it only has 200 and 400).'
        if 'responses' not in method:
            return self.match_error(responses_error, request)
        potential_responses = [r for r in method['responses'].items()]
        random.shuffle(potential_responses)
        if len(potential_responses) == 0:
            return self.match_error(responses_error, request)
        response = potential_responses[0]
        headers = {}
        if 'headers' in response:
            # TODO: can't handle references yet, need to fix
            headers = {} # { k: (faker(v['schema'], v['schema'], 0) if 'schema' in v else '***') for k,v in headers.items() }
        statusCode = int(response[0] if response[0] != 'default' else 400)
        if ('content' not in response[1]) or len(response[1]['content'].items()) == 0:
            return Response(statusCode=statusCode, body="", headers=headers)
        mime_types = response[1]['content'].keys()
        if "application/json" in mime_types:
            content = response[1]['content']['application/json']
            if 'schema' not in content:
                return self.match_error('Could not find schema', request)
            schema = content['schema']
            to_fake = {
                **(change_ref(schema) if '$ref' in schema else change_refs(schema)),
                'definitions': { k: change_ref(cast(Reference, v)) if '$ref' in v else change_refs(cast(Schema, v)) for k,v in (match[name]['components']['schemas'].items() if name in match and 'components' in match[name] and 'schemas' in match[name]['components'] else [])}
            }
            bodyAsJson = faker(to_fake, to_fake, 0)
            return Response(
                statusCode=statusCode,
                body=json.dumps(bodyAsJson),
                bodyAsJson=bodyAsJson,
                # TODO: can this be accomplished without a cast?
                headers=cast(Mapping[str, Union[str, Sequence[str]]], { **headers, "Content-Type": "application/json" })
            )
        if "text/plain" in mime_types:
            return Response(
                statusCode=statusCode,
                body=fkr.sentence(),
                # TODO: can this be accomplished without a cast?
                headers=cast(Mapping[str, Union[str, Sequence[str]]], { **headers, "Content-Type": "text/plain" })
            )
        return self.match_error('Could not produce content for these mime types %s' % str(mime_types), request)



import abc
from meeshkan.schemabuilder.result import BuildResult
from openapi_typed import OpenAPIObject
from typing import Union, Awaitable
from typing import Awaitable


class AbstractSink(abc.ABC):
    """Abstract sink for build results.
    """

    @abc.abstractmethod
    def push(self, result: BuildResult) -> Union[Awaitable[None], None]:
        """Push build result to the sink.

        Arguments:
            result {BuildResult} -- Build result object.

        Returns:
            Union[Awaitable[None], None] -- Awaitable or None.
        """
        raise NotImplementedError("Not implemented")

    @abc.abstractmethod
    def flush(self) -> Union[Awaitable[None], None]:
        """Flush the build results.

        Returns:
            Union[Awaitable[None], None] -- Awaitable or None
        """
        raise NotImplementedError("Not implemented")



"""Code for working with OpenAPI paths, e.g., matching request path to an OpenAPI endpoint with parameter."""
import re
import random
import string
from deepdiff import DeepDiff
from typing import cast, Pattern, Optional, Tuple, Mapping, Any
from typing_extensions import TypedDict

from openapi_typed import PathItem, Paths, Operation, Responses

# Pattern to match to in the escaped path string
# Search for occurrences such as "{id}" or "{key-name}"
PATH_PARAMETER_PATTERN = r"""\\{([\w-]+)\\}"""

RequestPathParameters = Mapping[str, str]


def _dumb_match_to_path(request_path: str, paths: Paths, request_method: str, operation_candidate: Operation) -> Tuple[str, Optional[str], Optional[Mapping[str, Any]]]:
    """An overly-simplistic, highly-experimental, probably-broken
       function that answers the question "does this path exist in the spec already?"
       For example, if there is /v1/clients/jane and we get
       /v1/clients/john, it should be smart enough to realize that the path is actually
       /v1/clients/{identifier}.

       We do this by using the `deepdiff` library and returning a match
       if there is enough "stuff" that doesn't diverge between two paths that
       could theoretically be the same when doing pattern matching.

       In the future, "machine learning" can take care of this... yeah...
       machine learning... the future... ooohhh....
    """

    def could_these_two_paths_possibly_represent_the_same_underlying_path(a: str, b: str) -> bool:
        """Asks the question "Could these two paths theoretically be the same?
        Uses some brutish heuristics to answer it.

        Arguments:
            a {str} -- A path
            b {str} -- A path
        
        Returns:
            bool -- Could they possibly be the same underlying path?
        """
        if a is b:
            return True
        a_spl = [x for x in a.split("/") if x != '']
        b_spl = [x for x in b.split("/") if x != '']
        # if the two lengths aren't equal, return false
        if len(a_spl) != len(b_spl):
            return False
        # if the first part of the path is not equal, return false
        # APIs almost never have a top-level wildcard
        if a_spl[0] != b_spl[0]:
            return False
        # if there are two successive non-matches, return false
        for x in range(len(a_spl) - 1):
            if a_spl[x] != b_spl[x]:
                if a_spl[x + 1] != b_spl[x + 1]:
                    return False
        # our naive test has passed, they could be theoretically similar...
        return True

    def combine_paths_into_single_path(path0: str, path1: str) -> str:
        """Combine two paths into a single path, adding wildcards when needed.

        Arguments:
            path0 {str} -- One path to combine
            path1 {str} -- Two paths to combine
        
        Returns:
            str -- A combined path
        """
        
        path0_split = [x for x in path0.split('/') if x != '']
        path1_split = [x for x in path1.split('/') if x != '']
        if len(path0_split) != len(path1_split):
            raise ValueError("Incorrect use of path combination function")
        return '/' + '/'.join([a if a == b else a if a[0] == '{' else b if b[0] == '{' else '{%s}' % ''.join(random.choices(string.ascii_lowercase, k=8)) for a, b in zip(path0_split, path1_split)])

    theoretically_plausible_paths = [path for path in paths.keys() if could_these_two_paths_possibly_represent_the_same_underlying_path(path, request_path)]
    for path in theoretically_plausible_paths:
        operation = cast(Operation, paths[path][request_method])
        potential_conflicts = set(cast(Responses, operation['responses']).keys()).intersection(set(cast(Responses, operation_candidate['responses']).keys()))
        new_path = combine_paths_into_single_path(path, request_path)
        if len(potential_conflicts) != 0:
            irreconcilably_different = False
            for potential_conflict in potential_conflicts:
                # we look to see if potential conflicts resemble each other
                # enough to call them the same thing
                # if not, we deem them as irreconcilably different and give up
                incumbent_potential_conflict = operation['responses'][potential_conflict]
                candidate_potential_conflict = operation_candidate['responses'][potential_conflict]
                diff = DeepDiff(incumbent_potential_conflict, candidate_potential_conflict)
                # we don't care about value differences
                # if there are too many differences based on my totally aribtrary standard,
                # we deem the two to be irreconcilably different
                if 'values_changed' in diff.keys():
                    del diff['values_changed']
                # If there are more than five things that are different between the specs,
                # we deem them as irreconcilably different.
                # This is based on the Helsinki Standard of Irreconcilable Differences Between JSON Objects,
                # ISO-2912432, which has been passed down as an oral traditoin since the dawn of time.
                if sum([len(x) for x in diff.values()], 0) > 5:
                    irreconcilably_different = True
                    break
            if irreconcilably_different:
                continue
        # if it has gotten this far, all potential
        # conflicts can be resolved, so we merge
        matches = _match_to_path(request_path, new_path)
        if matches is None:
            raise ValueError('Algorithm conflict for path matching - got a match for %s %s, but then returned match === None' % (request_path, new_path))
        # TODO: what if there is more than 1 theoretically plausible path?
        # in practice unlikely but possible if, for example, the spec has two conflicting
        # paths already in it
        return (new_path, path, matches)
    return (request_path, None, None)

def _match_to_path(request_path: str, path: str) -> Optional[Mapping[str, Any]]:
    """Match a request path such as "/pets/32" to a variable path such as "/pets/{petId}".

    Arguments:
        request_path {str} -- Request path such as /pets/32
        path {str} -- Path name in OpenAPI format such as /pets/{id}

    Returns:
        Optional[Mapping[str, Any]] -- None if the paths do not match. Otherwise, return a dictionary of parameter names to values (for example: { 'petId': '32' })
    """
    # this function won't work if
    # requests paths have a trailing slash, so we remove it
    if request_path[-1] == '/':
        request_path = request_path[:-1]
    path_as_regex, parameter_names = path_to_regex(path)
    match = path_as_regex.match(request_path)

    if match is None:
        return None

    captures = match.groups()

    assert len(parameter_names) == len(
        captures), "Expected the number of parameter names in the path to match the number of captured parameter values"

    return {parameter_name: parameter_value for parameter_name, parameter_value in zip(parameter_names, captures)}


class MatchingPath(TypedDict):
  path: PathItem
  param_mapping: Mapping[str, Any]
  pathname_with_wildcard: str
  pathname_to_be_replaced_with_wildcard: Optional[str]

def find_matching_path(request_path: str, paths: Paths, request_method: str, operation_candidate: Operation) -> Optional[MatchingPath]:
    """Find path that matches the request path.

    Arguments:
        request_path {str} -- Request path.
        paths {Paths} -- OpenAPI Paths object.
        request_method {str} -- The method (ie get, post, etc)
        operation_candidate {Operation} -- An operation to match against other paths.

    Returns:
        Optional[MatchingPath] -- The matching path (if any)
    """

    # First pass - we find an explicit match
    # Second pass - we construct a match if it looks like
    # two paths are in fact similar.
    for fn in [
        lambda p, pi: (request_path, None, _match_to_path(request_path=request_path, path=p)),
        lambda p, pi: _dumb_match_to_path(request_path, paths, request_method, operation_candidate)]:
        for path, path_item in paths.items():
            pathname_with_wildcard, pathname_to_be_replaced_with_wildcard, path_match = fn(path, path_item)

            if path_match is None:
                continue
            return {'path': path_item, 'param_mapping': path_match, 'pathname_with_wildcard': pathname_with_wildcard, 'pathname_to_be_replaced_with_wildcard': pathname_to_be_replaced_with_wildcard }

    return None


PATH_PARAMETER_REGEX = r"""(\w+)"""


def path_to_regex(path: str) -> Tuple[Pattern[str], Tuple[str]]:
    """Convert an OpenAPI path such as "/pets/{id}" to a regular expression. The returned regular expression
    contains capturing groups for path parameters. Parameter names are returned in the second tuple.

    Arguments:
        path {str} -- [description]

    Returns:
        {} -- Tuple containing (1) pattern for path with parameters replaced by regular expressions, and (2) list of parameters with names.
    """

    # Work on string whose regex characters are escaped ("/"" becomes "//" etc.)
    # This makes it easier to replace matches with regular expressions.
    # For example: /pets/{id} becomes \/pets\/\{id\}
    escaped_path = re.escape(path)

    param_names = ()  # type: Tuple[str]

    for match in re.finditer(PATH_PARAMETER_PATTERN, escaped_path):
        full_match = match.group(0)
        param_name = match.group(1)

        param_names = param_names + (param_name, )

        escaped_path = escaped_path.replace(full_match, PATH_PARAMETER_REGEX)

    regex_pattern = re.compile(r'^' + escaped_path + r'(?:\?|#|$)')

    return (regex_pattern, param_names)



from meeshkan.schemabuilder.paths import path_to_regex, find_matching_path
from hamcrest import *
from ..util import petstore_schema

PETSTORE_SCHEMA = petstore_schema()


def test_path_to_regex():
    as_regex, parameters = path_to_regex('/pets/{id}')

    # assert_that(as_regex.pattern, is_('^\\/pets\\/(\\w+)$'))
    # assert_that(as_regex.pattern, is_(r"""^\/pets\/(\w+)$"""))

    assert_that("/pets/32", matches_regexp(as_regex))
    assert_that("/pets/32?id=3", matches_regexp(as_regex))
    assert_that("/pets/32#reference", matches_regexp(as_regex))

    assert_that("/pets/32/", not_(matches_regexp(as_regex)))

    assert parameters == ('id', )


def test_path_to_regex_with_multiple_params():
    as_regex, _ = path_to_regex(
        '/pets/{id}/items/{name}')

    assert_that("/pets/32/items/car", matches_regexp(as_regex))


def test_match_paths():
    paths = PETSTORE_SCHEMA['paths']
    request_path = "/pets/32"

    match_result = find_matching_path(request_path, paths, "GET", {
        'responses': {}
    })

    assert match_result is not None

    expected_path_item = PETSTORE_SCHEMA['paths']['/pets/{petId}']
    path_item, parameters = match_result['path'], match_result['param_mapping']

    assert_that(path_item, equal_to(expected_path_item))
    assert_that(parameters, has_entry('petId', '32'))



"""Code for working with parameters."""

from functools import reduce
from meeshkan.schemabuilder.json_schema import to_const
from meeshkan.schemabuilder.update_mode import UpdateMode
from openapi_typed import Parameter, Reference, Schema
from typing import Sequence, Union, cast, Sequence, List, Mapping
from genson import SchemaBuilder  # type: ignore

SchemaParameters = Sequence[Union[Parameter, Reference]]

# TODO: this won't work for anything other than simple schema
# maybe make more robust?
def unnest(d: List[Schema]) -> List[Schema]:
    return cast(List[Schema], reduce(lambda a, b: [*a, *b],
        [[x] if 'oneOf' not in x else unnest(cast(List[Schema], x['oneOf'])) for x in d
    ], []))
    

# TODO: extended this for all paramaters, but need to verify that
# there is not some subtle difference between query and header

class ParamBuilder:
    def __init__(self, _in: str):
        if _in not in ['header', 'query']:
            raise ValueError('A parameter cannot have an `in` value of %s.' % _in)
        self._in = _in

    def build(self, params: Mapping[str, Union[str, Sequence[str]]], mode: UpdateMode) -> SchemaParameters:
        """Build a list of parameters from request parameters.

        Arguments:
            params {Mapping[str, str]} -- Key-value map of params parameters and values.

        Returns:
            SchemaParameters -- OpenAPI list of parameters.
        """
        existing = []
        # as a defualt, we set nothing new to required
        # can change later if there is a compelling reason
        return self.update(params, mode, existing, set_new_as_required=False)


    def build_param(self, name: str, value: Union[str, Sequence[str]], required: bool, mode: UpdateMode) -> Parameter:
        """Build a new OpenAPI compatible parameter definition from parameter
        name and value.

        Arguments:
            name {str} -- Parameter name.
            value {Any} -- Parameter value.
            required {bool} -- Whether the parameter should be marked as required.
        Returns:
            Parameter -- [description]
        """
        # TODO: this may not be accurate if there are duplicate parameters
        # which is allowed in http
        if isinstance(value, list) and len(value) == 1:
            value = value[0]
        else:
            value = value
        out = {}
        if (mode == UpdateMode.GEN):
            schema_builder = SchemaBuilder()
            schema_builder.add_object(value)
            out = schema_builder.to_schema()
        else:
            out = { 'oneOf': [to_const(value)]}
        schema = cast(Schema, out)
        return Parameter(name=name, schema=schema, required=required, **{'in': self._in})


    def _update_required(self, param: Parameter, required: bool) -> Parameter:
        if param['required'] == required:
            return param

        params = dict(**param)
        params.pop('required')
        return Parameter(**params, required=required)


    # TODO Fix types once openapi types are covariant
    def update(self, incoming_params: Mapping[str, Union[str, Sequence[str]]], mode: UpdateMode, existing: SchemaParameters, set_new_as_required=False) -> SchemaParameters:
        non_params: List[Parameter] = [
            param for param in existing if param['in'] != self._in]  # type: ignore

        params: List[Parameter] = [
            param for param in existing if param['in'] == self._in]  # type: ignore

        schema_params_names = frozenset(
            [param['name'] for param in params])

        request_param_names = frozenset(incoming_params.keys())

        # Parameters in request but not in schema
        new_param_names = request_param_names.difference(
            schema_params_names)

        # Parameters in schema but not in request
        missing_param_names = schema_params_names.difference(
            request_param_names)

        # Parameters in schema and in request
        shared_param_names = request_param_names.intersection(
            schema_params_names)

        new_params = [self.build_param(name=param_name,
                                            value=value,
                                            required=set_new_as_required,
                                            mode=mode)
                            for param_name in new_param_names
                            for value in (incoming_params[param_name],)]

        # TODO Update shared incoming_params parameter schema
        shared_params = [
            param if (mode == UpdateMode.GEN) else {
                # use constants
                **param,
                'schema': {'oneOf': unnest([param['schema'], to_const(incoming_params[param['name']])])}
            } for param in params if param['name'] in shared_param_names]

        missing_params = [
            self._update_required(param, required=False) for param in params if param['name'] in missing_param_names]

        # TODO: make sure cast is valid
        updated_params = [
            *new_params,
            *shared_params,
            *missing_params,
            *non_params
        ]
        return cast(SchemaParameters, updated_params)



import logging
import os

import click
import tornado.ioloop
import yaml
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.web import Application

from .utils.routing import PathRouting, HeaderRouting, Routing
from .admin.views import StorageView
from .proxy.proxy import RecordProxy
from .server.callbacks import callback_manager
from .server.response_matcher import ResponseMatcher
from .server.views import MockServerView
from .utils.data_callback import RequestLoggingCallback

LOG_CONFIG = os.path.join(os.path.dirname(__file__), 'logging.yaml')
logger = logging.getLogger(__name__)

def start_admin(port):
    app = Application([
        (r'/admin/storage', StorageView)
    ])
    http_server = HTTPServer(app)
    http_server.listen(port)
    logger.info('Starting admin endpont on http://localhost:%s/admin', port)



@click.command()
@click.option('--port', default="8000", help='Server port')
@click.option('--admin_port', default="8888", help='Admin server port')
@click.option('--log_dir', default="./logs", help='API calls logs direcotry')
@click.option('--header_routing', is_flag=True, help='Whether to use a header based routing to a target host')
@click.option('--schema_dir', default="./__unmock__", help='Directory with OpenAPI schemas')
def record(port, admin_port, log_dir, header_routing, schema_dir):
    """
    Record http traffic to http-types format.
    """
    start_admin(admin_port)
    logger.info('Starting Meeshkan proxy in recording mode on http://localhost:%s', port)
    with RequestLoggingCallback(recording=True, log_dir=log_dir, schema_dir=schema_dir) as callback:
        server = RecordProxy(callback, HeaderRouting() if header_routing else PathRouting())
        server.listen(port)
        tornado.ioloop.IOLoop.instance().start()

class MeeshkanApplication(Application):
    response_matcher: ResponseMatcher
    router: Routing

def make_mocking_app(callback_path, schema_dir, router):
    app = MeeshkanApplication([
        (r'/.*', MockServerView)
    ])
    callback_manager.load(callback_path)

    app.response_matcher = ResponseMatcher(schema_dir)
    app.router = router
    return app

@click.command()
@click.option('--callback_path', default="./callbacks", help='Directory with configured callbacks')
@click.option('--admin_port', default="8888", help='Admin server port')
@click.option('--port', default="8000", help='Server port')
@click.option('-s', '--schema_dir', default="./__unmock__", help='Directory with OpenAPI schemas')
@click.option('--header_routing', is_flag=True, help='Whether to use a path based routing to a target host')
def mock(callback_path, admin_port, port, schema_dir, header_routing):
    start_admin(admin_port)
    app = make_mocking_app(callback_path, schema_dir, HeaderRouting() if header_routing else PathRouting())
    http_server = HTTPServer(app)
    http_server.listen(port)
    logger.info('Mock server is listening on http://localhost:%s', port)
    IOLoop.current().start()



from meeshkan.schemabuilder.json_schema import to_json_schema, to_openapi_json_schema
from meeshkan.schemabuilder.update_mode import UpdateMode
from ..util import read_recordings_as_dict
import json
from hamcrest import *

request_samples = read_recordings_as_dict()
object_sample = json.loads(request_samples[0]['response']['body'])


test_objects = [
    [
        {'fork': True}
    ],
    [
        {}
    ]
]


def test_to_schema_from_array():
    schema = to_json_schema(object_sample, UpdateMode.GEN, schema=None)
    assert_that(schema, has_key("$schema"))
    assert_that(schema, has_entry(
        "$schema", starts_with('http://json-schema.org/schema#')))
    assert_that(schema, has_entry("type", "array"))
    assert_that(schema, has_entry('items', instance_of(dict)))


def test_schema_with_array():
    schema = to_json_schema(test_objects[0], UpdateMode.GEN, schema=None)
    assert_that(schema, has_entry("items", instance_of(dict)))
    items = schema['items']
    assert isinstance(items, dict)  # typeguard
    assert_that(items, has_entry('required', ['fork']))
    assert_that(items, has_entry('properties', instance_of(dict)))
    properties = items['properties']
    assert_that(properties, has_entries(fork=instance_of(dict)))
    forks = properties['fork']
    assert_that(forks, has_entries(type="boolean"))


def test_updating_schema_removes_required():
    schema = to_json_schema(test_objects[0], UpdateMode.GEN, schema=None)
    schema = to_json_schema(test_objects[1], UpdateMode.GEN, schema=schema)
    items = schema['items']
    assert isinstance(items, dict)  # typeguard
    assert_that(items, has_entry('required', []))
    assert_that(items, has_entry('properties', has_key("fork")))


def test_openapi_compatible_schema():
    schema = to_openapi_json_schema(test_objects[0], UpdateMode.GEN, schema=None)
    assert_that(schema, not_(has_key("$schema")))



from typing import AsyncIterable, Tuple, TextIO
from http_types.types import HttpExchange
from http_types.utils import HttpExchangeReader
from .abstract import AbstractSource
import asyncio


class FileSource(AbstractSource):

    def __init__(self, input_file: TextIO):
        self.input_file = input_file

    async def start(self, loop: asyncio.events.AbstractEventLoop) -> Tuple[AsyncIterable[HttpExchange], None]:
        # print(self.input_file)

        async def read():
            for line in self.input_file:
                yield HttpExchangeReader.from_json(line)

        return read(), None

    def shutdown(self):
        pass



"""Code for writing builder results to file system.
"""
from openapi_typed import OpenAPIObject
from .result import BuildResult
import yaml
from ..logger import get as getLogger
from typing import cast
from pathlib import Path

LOGGER = getLogger(__name__)


OPENAPI_FILENAME = 'openapi.yaml'


def _read_openapi(file: Path) -> OpenAPIObject:
    """Read an OpenAPI YAML from file.

    Arguments:
        file {Path} -- Path to OpenAPI YAML file.

    Returns:
        OpenAPIObject -- OpenAPI object.
    """
    with file.open('rb') as f:
        return cast(OpenAPIObject, yaml.safe_load(f))


def read_directory(directory: str) -> BuildResult:
    """Read BuildResult from directory.

    Arguments:
        directory {str} -- Directory to read, possibly relative.

    Raises:
        FileNotFoundError: If directory is not an existing directory.

    Returns:
        BuildResult -- Build result.
    """
    path = _resolve_path(directory)

    if not path.is_dir():
        raise FileNotFoundError("Cannot read from {}".format(str(path)))

    openapi_path = path / OPENAPI_FILENAME
    openapi_object = _read_openapi(openapi_path)

    return BuildResult(openapi=openapi_object)


def _ensure_dir_exists(path: Path) -> None:
    """Ensure directory at `path` exists. Does NOT create parent directories.

    Arguments:
        path {Path} -- Path to the directory to create.

    Raises:
        FileNotFoundError -- If the directory does not exist and its parents do not exist.
    """
    path.mkdir(parents=False, exist_ok=True)


def _resolve_path(directory: str) -> Path:
    """Return directory as absolute Path.

    Arguments:
        directory {str} -- Path to directory, possibly relative.

    Returns:
        Path -- Absolute Path.
    """
    path = Path(directory)
    return path.resolve()


def write_build_result(directory: str, result: BuildResult) -> None:
    """Write builder result to a directory.

    Arguments:
        directory {str} -- Directory where to write results, possibly relative.
        result {BuildResult} -- Builder result.
    """
    output_dir_path = _resolve_path(directory)
    LOGGER.info("Writing to folder %s.", str(output_dir_path))

    _ensure_dir_exists(output_dir_path)

    openapi_output = output_dir_path / OPENAPI_FILENAME

    openapi_object = result['openapi']

    schema_yaml = cast(str, yaml.safe_dump(openapi_object))

    with openapi_output.open('wb') as f:
        LOGGER.debug("Writing to: %s\n", str(openapi_output))
        f.write(schema_yaml.encode(encoding="utf-8"))



from meeshkan.schemabuilder.result import BuildResult
from meeshkan.schemabuilder.writer import write_build_result
from .abstract import AbstractSink


class FileSystemSink(AbstractSink):

    def __init__(self, out_dir: str):
        self.out_dir = out_dir
        self.result = None

    def push(self, result: BuildResult) -> None:
        self.result = result

    def flush(self) -> None:
        if self.result is None:
            return
        write_build_result(self.out_dir, result=self.result)



from openapi_typed import Server
from http_types import Request
from typing import List, Optional, Sequence
from urllib.parse import urlparse


def normalize_path_if_matches(request: Request, servers: Sequence[Server]) -> Optional[str]:
    """Check if a request matches a list of server definitions.
    If matches, return the request path normalized by the server URL's basepath.

    For example: for matching server URL `https://petstore.swagger.io/v1` and request URL
    `https://petstore.swagger.io/v1/pets`, return `/pets`.

    Arguments:
        servers {List[Server]} -- List of schema Server definitions.
        request {Request} -- HTTP request.

    Returns:
        Optional[str] -- None if request does not match any existing servers. Request path normalized by server basepath otherwise. 
    """

    for server in servers:
        server_url = urlparse(server['url'])
        if server_url.scheme != request['protocol']:
            continue

        if server_url.netloc != request['host']:
            continue

        if not request['pathname'].startswith(server_url.path):
            continue

        return request['pathname'][len(server_url.path):]

    return None



import copy
from meeshkan.schemabuilder.update_mode import UpdateMode
import json
import re
from http_types import HttpExchange, Request, Response, RequestBuilder
from tests.schemabuilder.paths_test import PETSTORE_SCHEMA
from meeshkan.schemabuilder import build_schema_batch, update_openapi
from meeshkan.schemabuilder.builder import BASE_SCHEMA
from meeshkan.schemabuilder.schema import validate_openapi_object
from ..util import petstore_schema, read_recordings_as_request_response, POKEAPI_RECORDINGS_PATH
from openapi_typed import OpenAPIObject, Operation, PathItem, Response, Schema
from typeguard import check_type
import pytest
from typing import cast
from hamcrest import *

requests = read_recordings_as_request_response()
pokeapi_requests = read_recordings_as_request_response(POKEAPI_RECORDINGS_PATH)

expected_schema = copy.deepcopy(BASE_SCHEMA)

expected_schema['paths'] = {
    '/user/repos': PathItem(
        get=Operation(
            responses={
                '200': Response(description="description"),
                '403': Response(description="description")
            }
        )
    )
}

PETSTORE_SCHEMA = petstore_schema()


class TestSchema:
    schema = build_schema_batch(requests, UpdateMode.GEN)
    pokeapi_schema = build_schema_batch(pokeapi_requests, UpdateMode.GEN)

    def response_schema(self):

        return cast(Schema, self.schema['paths']['/user/repos']
                    ['get']['responses']['200']
                    ['content']['application/json']['schema'])  # type: ignore

    def test_schema_valid(self):
        validate_openapi_object(self.schema)

    def test_schema_typechecks(self):
        check_type('schema', self.schema, OpenAPIObject)

    def test_schema_keys(self):
        assert 'openapi' in self.schema
        assert 'info' in self.schema
        assert 'paths' in self.schema

    def test_paths(self):
        paths = self.schema['paths']
        assert '/user/repos' in paths

    def test_get_operation(self):
        op = self.schema['paths']['/user/repos']['get']
        assert 'responses' in op

    def test_get_operation_responses(self):
        responses = self.schema['paths']['/user/repos']['get']['responses']
        assert '200' in responses
        assert '403' in responses

    def test_200_response(self):
        response = self.schema['paths']['/user/repos']['get']['responses']['200']
        # TODO Typeguard
        response = cast(Response, response)
        assert response['headers'] == {}
        assert response['links'] == {}
        assert 'application/json' in response['content']
        media_type = response['content']['application/json']
        assert 'schema' in media_type

    def test_repos_content(self):
        content = cast(
            Response, self.schema['paths']['/user/repos']['get']['responses']['200'])['content']
        assert_that(content, has_key("application/json"))
        json_content = content['application/json']
        assert_that(json_content, has_key("schema"))
        schema = json_content['schema']
        assert_that(schema, has_entry("type", "array"))
        # TODO Typeguard for Reference
        schema = cast(Schema, schema)
        assert_that(schema, has_entry("items", instance_of(dict)))

    def test_items(self):
        schema = self.response_schema()
        items = schema['items']
        assert isinstance(items, dict)  # typeguard
        assert_that(items, has_entry('properties', instance_of(dict)))
        properties = items['properties']
        assert_that(properties, has_entry('clone_url',
                                          equal_to({'type': 'string'})))

    def test_servers(self):
        servers = self.schema['servers']
        assert 1 == len(servers)
        assert 'http://api.github.com' == servers[0]['url']

    def test_pokeapi_schema_valid(self):
        # this should conflate to
        # /pokemon
        # /abilities/*
        # /types/*
        # /pokemon/*
        # meaning that it should recognize wildcards
        # from all these paths
        paths = self.pokeapi_schema['paths'].keys()
        assert 4 == len(paths)
        assert_that(paths, has_item("/v2/pokemon/"))
        assert_that(paths, has_item(
            matches_regexp(r'\/v2\/pokemon\/\{[\w]+\}')))
        assert_that(paths, has_item(
            matches_regexp(r'\/v2\/type\/\{[\w]+\}')))
        assert_that(paths, has_item(
            matches_regexp(r'\/v2\/ability\/\{[\w]+\}')))


class TestPetstoreSchemaUpdate:

    get_pets_req = RequestBuilder.from_url(
        "http://petstore.swagger.io/v1/pets")
    get_one_pet_req = RequestBuilder.from_url(
        "http://petstore.swagger.io/v1/pets/32")

    res = Response(body="", statusCode=200, headers={})

    get_one_pet_exchange = HttpExchange(request=get_one_pet_req, response=res)
    get_pets_exchange = HttpExchange(request=get_pets_req, response=res)

    orig_schema_paths = list(PETSTORE_SCHEMA['paths'].keys())

    def test_update_respects_path_parameter(self):
        updated_schema = update_openapi(
            PETSTORE_SCHEMA, self.get_one_pet_exchange, UpdateMode.GEN)
        updated_schema_paths = list(updated_schema['paths'].keys())
        assert_that(updated_schema_paths, is_(self.orig_schema_paths))

    def test_update_respects_basepath(self):
        updated_schema = update_openapi(
            PETSTORE_SCHEMA, self.get_pets_exchange, UpdateMode.GEN)
        updated_schema_paths = list(updated_schema['paths'].keys())
        assert_that(updated_schema_paths, is_(self.orig_schema_paths))


class TestQueryParameters:

    req = RequestBuilder.from_url(
        url="https://petstore.swagger.io/v1/pets/32?id=1&car=ferrari")
    res = Response(body="", statusCode=200, headers={})
    exchange = HttpExchange(request=req, response=res)

    req_wo_query = RequestBuilder.from_url(
        url="https://petstore.swagger.io/v1/pets/32")
    res = Response(body="", statusCode=200, headers={})
    exchange_wo_query = HttpExchange(request=req_wo_query, response=res)

    expected_path_name = "/v1/pets/32"

    def test_build_with_query(self):
        schema = build_schema_batch([self.exchange], UpdateMode.GEN)

        assert_that(list(schema['paths'].keys()),
                    is_([self.expected_path_name]))

        operation = schema['paths'][self.expected_path_name]['get']

        assert_that(operation, has_key('parameters'))

        parameters = operation['parameters']

        assert_that(parameters, has_length(2))

        parameter_names = [param['name'] for param in parameters]

        assert_that(set(parameter_names), is_(set(['id', 'car'])))

    def test_schema_update_with_query(self):
        schema = build_schema_batch([self.exchange_wo_query], UpdateMode.GEN)

        operation = schema['paths'][self.expected_path_name]['get']
        assert_that(operation, has_entry('parameters', []))

        updated_schema = update_openapi(schema, self.exchange, UpdateMode.GEN)

        operation = updated_schema['paths'][self.expected_path_name]['get']
        assert_that(operation, has_entry('parameters', has_length(2)))

        first_query_param = operation['parameters'][0]

        assert_that(first_query_param, has_entry("required", False))


class TestSchemaTextBody:
    request = RequestBuilder.from_url("https://example.com/v1")
    response = Response(statusCode=200, body="Hello World", headers={})
    exchange: HttpExchange = {'request': request, 'response': response}

    def test_build_string_body(self):
        schema = build_schema_batch([self.exchange], UpdateMode.GEN)
        response_content = schema['paths']['/v1']['get']['responses']['200']['content']

        assert_that(response_content, has_key("text/plain"))

        media_type = response_content["text/plain"]

        assert_that(media_type, has_entry(
            "schema", has_entry("type", "string")))

class TestShemaInReplayMode:

    def test_schema_in_replay_mode(self):
        reqs = []
        with open('tests/server/mock/callbacks/opbank/recordings/recording.jsonl','r') as rr: 
            reqs = rr.read().split('\n')

        reqs = [json.loads(r) for r in reqs if r != '']
        r = build_schema_batch(reqs, UpdateMode.REPLAY)
        # this schema has four recordings, of which two correspond to /v1/payments/confirm
        assert 2 == len(r['paths']['/v1/payments/confirm']['post']['responses']['201']['content']['application/json']['schema']['oneOf'])